#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/3/20 4:44 下午
# @Author  : Han Yu    Dang Haoran    Cai Chao
# @File    : two_qubit_gate_rules.py

import numpy as np

from QuICT.core.gate import *

"""
the file describe transform rules between two kinds of 2-qubit gates.
"""
"""
transform CX gate into others
"""


def cx2cy_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        S & targs[1]
        CY & targs
        S_dagger & targs[1]
    return gates


def cx2cz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[1]
        CZ & targs
        H & targs[1]
    return gates


def cx2ch_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        T & targs[1]
        Rx(np.pi / 2) & targs[1]
        CH & targs
        Rx(-np.pi / 2) & targs[1]
        T_dagger & targs[1]
    return gates


def cx2crz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 4) & targs[0]
        Rz(np.pi / 2) & targs[0]
        H & targs[1]
        CRz(np.pi) & targs
        H & targs[1]
    return gates


def cx2fsim_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[1]
        FSim(0, np.pi) & targs
        H & targs[1]
    return gates


def cx2rxx_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(7 * np.pi / 4) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
    return gates


def cx2ryy_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(3 * np.pi / 4) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[1]
        Ryy(np.pi / 2) & targs
        Rx(np.pi / 2) & targs[1]
        Rx(np.pi / 2) & targs[0]
        H & targs[1]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
    return gates


def cx2rzz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(7 * np.pi / 4) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        H & targs[0]
        H & targs[1]
        Rzz(np.pi / 2) & targs
        H & targs[1]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
    return gates


"""
transform CY gate into others
"""


def cy2cx_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        S_dagger & targs[1]
        CX & [targs[0], targs[1]]
        S & targs[1]
    return gates


def cy2cz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        S_dagger & targs[1]
        H & targs[1]
        CZ & targs
        H & targs[1]
        S & targs[1]
    return gates


def cy2ch_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        T_dagger & targs[1]
        Rx(np.pi / 2) & targs[1]
        CH & targs
        Rx(-np.pi / 2) & targs[1]
        T & targs[1]
    return gates


def cy2crz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 4) & targs[0]
        Rz(np.pi / 2) & targs[0]
        S_dagger & targs[1]
        H & targs[1]
        CRz(np.pi) & targs
        H & targs[1]
        S & targs[1]
    return gates


def cy2fsim_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        S_dagger & targs[1]
        H & targs[1]
        FSim(0, np.pi) & targs
        H & targs[1]
        S & targs[1]
    return gates


def cy2rxx_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(7 * np.pi / 4) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        S_dagger & targs[1]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
        S & targs[1]
    return gates


def cy2ryy_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(3 * np.pi / 4) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[0]
        Ryy(np.pi / 2) & targs
        Rx(np.pi) & targs[1]
        H & targs[1]
        Rx(np.pi / 2) & targs[0]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
    return gates


def cy2rzz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        H & targs[0]
        Rx(np.pi / 2) & targs[1]
        Rzz(np.pi / 2) & targs
        H & targs[1]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
        S & targs[1]
    return gates


"""
transform CZ gate into others
"""


def cz2cx_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[1]
        CX & [targs[0], targs[1]]
        H & targs[1]
    return gates


def cz2cy_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[1]
        S & targs[1]
        CY & targs
        S_dagger & targs[1]
        H & targs[1]
    return gates


def cz2ch_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(-np.pi / 4) & targs[1]
        S_dagger & targs[1]
        CH & targs
        S & targs[1]
        Rx(np.pi / 4) & targs[1]
    return gates


def cz2crz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 4) & targs[0]
        Rz(np.pi / 2) & targs[0]
        CRz(np.pi) & targs
    return gates


def cz2fsim_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        FSim(0, np.pi) & targs
    return gates


def cz2rxx_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(7 * np.pi / 4) & targs[0]
        H & targs[1]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
        H & targs[1]
    return gates


def cz2ryy_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 4) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[0]
        Rz(-np.pi / 2) & targs[1]
        Rx(-np.pi / 2) & targs[1]
        Ryy(np.pi / 2) & targs
        Rx(np.pi / 2) & targs[1]
        Rx(np.pi / 2) & targs[0]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
    return gates


def cz2rzz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(7 * np.pi / 4) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        H & targs[0]
        Rz(-np.pi / 2) & targs[1]
        Rzz(np.pi / 2) & targs
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
    return gates


"""
transform CRz gate into others
"""


def crz2cx_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        """decomposition1:"""
        Rz(theta / 2) & targs[1]
        CX & targs
        Rz(-theta / 2) & targs[1]
        CX & targs
        """decomposition2:
        U1(theta / 2) & targs[1]
        CX & targs
        U1(-theta / 2) & targs[1]
        CX & targs
        """
    return gates


def crz2cy_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 4) & targs[0]
        Rz((theta + np.pi) / 2) & targs[1]
        CY & targs
        Rz(-theta / 2) & targs[1]
        CY & targs
        S_dagger & targs[1]
    return gates


def crz2cz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rz(theta / 2) & targs[1]
        H & targs[1]
        CZ & targs
        Rx(-theta / 2) & targs[1]
        CZ & targs
        H & targs[1]
    return gates


def crz2ch_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 8) & targs[0]
        Rz(theta / 2 + np.pi / 4) & targs[1]
        Rx(np.pi / 2) & targs[1]
        CH & targs
        Ry(theta / 2) & targs[1]
        CH & targs
        Rx(-np.pi / 2) & targs[1]
        T_dagger & targs[1]
    return gates


def crz2fsim_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rz(theta / 2) & targs[1]
        H & targs[1]
        FSim(0, np.pi) & targs
        Rx(-theta / 2) & targs[1]
        FSim(0, np.pi) & targs
        H & targs[1]
    return gates


def crz2rxx_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(3 * np.pi / 2) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        U1(theta / 2) & targs[1]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        U1(-theta / 2) & targs[1]
        Rx(-np.pi / 2) & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
    return gates


def crz2ryy_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi) & targs[0]
        Rz(theta / 2 + np.pi / 2) & targs[1]
        Ry(np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[0]
        Ryy(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
        Rx(np.pi - theta / 2) & targs[1]
        H & targs[1]
        Ryy(np.pi / 2) & targs
        Rx(np.pi / 2) & targs[1]
        Rx(np.pi / 2) & targs[0]
        H & targs[1]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
    return gates


def crz2rzz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rz(theta / 2 + np.pi / 2) & targs[1]
        Rx(np.pi / 2) & targs[1]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        H & targs[0]
        Rzz(np.pi / 2) & targs
        Rz(-np.pi / 2) & targs[0]
        Rx(-theta / 2) & targs[1]
        Rz(-np.pi / 2) & targs[1]
        Rzz(np.pi / 2) & targs
        H & targs[1]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
    return gates


"""
transform CH gate into others
"""


def ch2cx_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(-np.pi / 2) & targs[1]
        T_dagger & targs[1]
        CX & targs
        T & targs[1]
        Rx(np.pi / 2) & targs[1]
    return gates


def ch2cy_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(-np.pi / 2) & targs[1]
        Rz(np.pi / 4) & targs[1]
        CY & targs
        Rz(-np.pi / 4) & targs[1]
        Rx(np.pi / 2) & targs[1]
    return gates


def ch2cz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        S & targs[1]
        Rx(np.pi / 4) & targs[1]
        CZ & targs
        Rx(-np.pi / 4) & targs[1]
        S_dagger & targs[1]
    return gates


def ch2crz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 4) & targs[0]
        Rz(np.pi / 2) & targs[0]
        S & targs[1]
        Rx(np.pi / 4) & targs[1]
        CRz(np.pi) & targs
        Rx(-np.pi / 4) & targs[1]
        S_dagger & targs[1]
    return gates


def ch2fsim_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        S & targs[1]
        Rx(np.pi / 4) & targs[1]
        FSim(0, np.pi) & targs
        Rx(-np.pi / 4) & targs[1]
        S_dagger & targs[1]
    return gates


def ch2rxx_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(13 * np.pi / 8) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi) & targs[1]
        Ry(-np.pi / 4) & targs[1]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
        T & targs[1]
        Rx(np.pi / 2) & targs[1]
    return gates


def ch2ryy_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rz(np.pi / 4) & targs[1]
        Ry(np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[0]
        Ryy(np.pi / 2) & targs
        Rx(np.pi / 2) & targs[0]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
        Rx(np.pi / 4) & targs[1]
        S_dagger & targs[1]
    return gates


def ch2rzz_rule(gate):
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(3 * np.pi / 2) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        H & targs[0]
        Ry(np.pi / 4) & targs[1]
        Rx(-np.pi) & targs[1]
        H & targs[1]
        Rzz(np.pi / 2) & targs
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
        Rx(-np.pi / 4) & targs[1]
        S_dagger & targs[1]
    return gates


"""
transform Rxx gate into others
"""


def rxx2cx_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[0]
        H & targs[1]
        CX & targs
        Rz(theta) & targs[1]
        CX & targs
        H & targs[1]
        H & targs[0]
    return gates


def rxx2cy_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[0]
        H & targs[1]
        S & targs[1]
        CY & targs
        Rz(theta) & targs[1]
        CY & targs
        S_dagger & targs[1]
        H & targs[1]
        H & targs[0]
    return gates


def rxx2cz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[0]
        CZ & targs
        Rx(theta) & targs[1]
        CZ & targs
        H & targs[0]
    return gates


def rxx2ch_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[0]
        Rx(-np.pi / 4) & targs[1]
        S_dagger & targs[1]
        CH & targs
        Ry(-theta) & targs[1]
        CH & targs
        S & targs[1]
        Rx(np.pi / 4) & targs[1]
        H & targs[0]
    return gates


def rxx2crz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 2) & targs[0]
        H & targs[0]
        Rz(np.pi) & targs[0]
        CRz(np.pi) & targs
        Rx(theta) & targs[1]
        CRz(np.pi) & targs
        H & targs[0]
    return gates


def rxx2fsim_rule(gate):
    targs = gate.cargs + gate.targs
    theta = gate.pargs[0]
    gates = CompositeGate()
    with gates:
        H & targs[0]
        FSim(0, np.pi) & targs
        Rx(theta) & targs[1]
        FSim(0, np.pi) & targs
        H & targs[0]
    return gates


def rxx2ryy_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[0]
        Rx(-np.pi / 2) & targs[0]
        H & targs[1]
        Rx(-np.pi / 2) & targs[1]
        Ryy(theta) & targs
        Rx(np.pi / 2) & targs[1]
        H & targs[1]
        Rx(np.pi / 2) & targs[0]
        H & targs[0]
    return gates


def rxx2rzz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[0]
        H & targs[1]
        Rzz(theta) & targs
        H & targs[1]
        H & targs[0]
    return gates


"""
transform Ryy gate into others
"""


def ryy2cx_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        CX & targs
        Rz(theta) & targs[1]
        CX & targs
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
    return gates


def ryy2cy_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        S & targs[1]
        CY & targs
        Rz(theta) & targs[1]
        CY & targs
        S_dagger & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
    return gates


def ryy2cz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        H & targs[1]
        CZ & targs
        Rx(theta) & targs[1]
        CZ & targs
        H & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
    return gates


def ryy2ch_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(np.pi / 2) & targs[0]
        Ry(np.pi / 4) & targs[1]
        Rx(np.pi) & targs[1]
        CH & targs
        Ry(-theta) & targs[1]
        CH & targs
        Rx(-np.pi) & targs[1]
        Ry(-np.pi / 4) & targs[1]
        Rx(-np.pi / 2) & targs[0]
    return gates


def ryy2crz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        H & targs[1]
        Rz(np.pi) & targs[0]
        CRz(np.pi) & targs
        Rx(theta) & targs[1]
        CRz(np.pi) & targs
        H & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
    return gates


def ryy2fsim_rule(gate):
    targs = gate.cargs + gate.targs
    theta = gate.pargs[0]
    gates = CompositeGate()
    with gates:
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        H & targs[1]
        FSim(0, np.pi) & targs
        Rx(theta) & targs[1]
        FSim(0, np.pi) & targs
        H & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
    return gates


def ryy2rxx_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        """Decomposition1:"""
        GPhase(7 * np.pi / 2) & targs[0]
        Rz(-np.pi / 2) & targs[0]
        Rxx(np.pi / 2) & targs
        Rz(theta) & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        """Decomposition2:
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        H & targs[0]
        H & targs[1]
        Rxx(theta) & targs
        H & targs[1]
        H & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rx(-np.pi / 2) & targs[0]
        """
    return gates


def ryy2rzz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        Rzz(theta) & targs
        Rx(-np.pi / 2) & targs[1]
        Rx(-np.pi / 2) & targs[0]
    return gates


"""
transform Rzz gate into others
"""


def rzz2cx_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        CX & targs
        Rz(theta) & targs[1]
        CX & targs
    return gates


def rzz2cy_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        S & targs[1]
        CY & targs
        Rz(theta) & targs[1]
        CY & targs
        S_dagger & targs[1]
    return gates


def rzz2cz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        H & targs[1]
        CZ & targs
        Rx(theta) & targs[1]
        CZ & targs
        H & targs[1]
    return gates


def rzz2ch_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        T & targs[1]
        Rx(np.pi / 2) & targs[1]
        CH & targs
        Ry(-theta) & targs[1]
        CH & targs
        Rx(-np.pi / 2) & targs[1]
        T_dagger & targs[1]
    return gates


def rzz2crz_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(np.pi / 2) & targs[0]
        H & targs[1]
        Rz(np.pi) & targs[0]
        CRz(np.pi) & targs
        Rx(theta) & targs[1]
        CRz(np.pi) & targs
        H & targs[1]
    return gates


def rzz2fsim_rule(gate):
    targs = gate.cargs + gate.targs
    theta = gate.pargs[0]
    gates = CompositeGate()
    with gates:
        H & targs[1]
        FSim(0, np.pi) & targs
        Rx(theta) & targs[1]
        FSim(0, np.pi) & targs
        H & targs[1]
    return gates


def rzz2rxx_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(7 * np.pi / 2) & targs[0]
        Ry(np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        Rz(theta) & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
    return gates


def rzz2ryy_rule(gate):
    theta = gate.pargs[0]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Ryy(theta) & targs
        Rx(np.pi / 2) & targs[1]
        Rx(np.pi / 2) & targs[0]
    return gates


"""
transform Fsim gate into others
"""


def fsim2cx_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 4) & targs[0]
        CX & targs
        H & targs[0]
        CX & [targs[1], targs[0]]
        Rz(-theta) & targs[0]
        CX & [targs[1], targs[0]]
        Rz(theta) & targs[0]
        H & targs[0]
        CX & targs
        U1(-fai / 2) & targs[1]
        CX & targs
        U1(fai / 2) & targs[1]
        CX & targs
        Rz(-fai / 2) & targs[0]
    return gates


def fsim2cy_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 4 - np.pi / 4) & targs[0]
        S & targs[1]
        CY & targs
        H & targs[0]
        S & targs[0]
        CY & [targs[1], targs[0]]
        Rz(-theta) & targs[0]
        CY & [targs[1], targs[0]]
        Rz(theta - np.pi / 2) & targs[0]
        H & targs[0]
        CY & targs
        U1(-fai / 2) & targs[1]
        CY & targs
        U1(fai / 2) & targs[1]
        CY & targs
        S_dagger & targs[1]
        Rz(-fai / 2) & targs[0]
    return gates


def fsim2cz_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 4) & targs[0]
        H & targs[1]
        CZ & targs
        H & targs[1]
        CZ & [targs[1], targs[0]]
        Rx(-theta) & targs[0]
        CZ & [targs[1], targs[0]]
        Rx(theta) & targs[0]
        H & targs[1]
        CZ & targs
        Rx(-fai / 2) & targs[1]
        CZ & targs
        Rx(fai / 2) & targs[1]
        CZ & targs
        H & targs[1]
        Rz(-fai / 2) & targs[0]
    return gates


def fsim2ch_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 4) & targs[0]
        T & targs[1]
        Rx(np.pi / 2) & targs[1]
        CH & targs
        Rx(-np.pi / 2) & targs[1]
        T_dagger & targs[1]
        Rx(-np.pi / 4) & targs[0]
        S_dagger & targs[0]
        CH & [targs[1], targs[0]]
        Ry(theta) & targs[0]
        CH & [targs[1], targs[0]]
        S & targs[0]
        Rx(theta + np.pi / 4) & targs[0]
        T & targs[1]
        Rx(np.pi / 2) & targs[1]
        CH & targs
        Ry(fai / 2) & targs[1]
        CH & targs
        Ry(-fai / 2) & targs[1]
        CH & targs
        Rx(-np.pi / 2) & targs[1]
        T_dagger & targs[1]
        Rz(-fai / 2) & targs[0]
    return gates


def fsim2crz_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 4 - np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[0]
        H & targs[1]
        CRz(np.pi) & targs
        H & targs[1]
        Rz(np.pi / 2) & targs[1]
        CRz(np.pi) & [targs[1], targs[0]]
        Rz(np.pi / 2) & targs[1]
        Rx(-theta) & targs[0]
        CRz(np.pi) & [targs[1], targs[0]]
        Rx(theta) & targs[0]
        Rz(np.pi / 2) & targs[0]
        H & targs[1]
        CRz(np.pi) & targs
        Rz(np.pi / 2) & targs[0]
        Rx(-fai / 2) & targs[1]
        CRz(np.pi) & targs
        Rz(np.pi / 2) & targs[0]
        Rx(fai / 2) & targs[1]
        CRz(np.pi) & targs
        H & targs[1]
        Rz(-fai / 2) & targs[0]
    return gates


def fsim2rxx_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 2 - np.pi / 2) & targs[0]
        Rxx(theta) & targs
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        H & targs[0]
        H & targs[1]
        Rxx(theta) & targs
        H & targs[0]
        Rx(-np.pi) & targs[0]
        Rz(-np.pi / 2) & targs[0]
        H & targs[1]
        Rx(-np.pi) & targs[1]
        Ry(-fai / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        U1(fai / 2) & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Rxx(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
        Rz(-fai / 2) & targs[0]
    return gates


def fsim2ryy_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 4 + np.pi / 2) & targs[0]
        H & targs[0]
        H & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Rx(-np.pi / 2) & targs[1]
        Ryy(theta) & targs
        Rx(np.pi / 2) & targs[1]
        Rx(np.pi / 2) & targs[0]
        H & targs[1]
        H & targs[0]
        Ryy(theta) & targs
        Rz(np.pi / 2 + -fai / 2) & targs[1]
        Ry(np.pi / 2) & targs[0]
        Rz(np.pi / 2) & targs[0]
        Ryy(np.pi / 2) & targs
        Ry(-np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        H & targs[1]
        Rz(np.pi / 2 + fai / 2) & targs[1]
        Ryy(np.pi / 2) & targs
        Rx(np.pi / 2) & targs[1]
        H & targs[1]
        Rx(-np.pi / 2) & targs[0]
        Ry(-np.pi) & targs[0]
        Rz(-fai / 2) & targs[0]
    return gates


def fsim2rzz_rule(gate):
    theta = gate.pargs[0]
    fai = gate.pargs[1]
    targs = gate.cargs + gate.targs
    gates = CompositeGate()
    with gates:
        GPhase(-fai / 2 - np.pi / 2) & targs[0]
        H & targs[0]
        H & targs[1]
        Rzz(theta) & targs
        H & targs[1]
        H & targs[0]
        Rx(np.pi / 2) & targs[0]
        Rx(np.pi / 2) & targs[1]
        Rzz(theta) & targs
        Rx(-np.pi) & targs[0]
        Rz(-np.pi / 2) & targs[0]
        H & targs[0]
        Rx(-np.pi) & targs[1]
        Ry(-fai / 2) & targs[1]
        H & targs[1]
        Rzz(np.pi / 2) & targs
        Rz(-np.pi / 2) & targs[0]
        H & targs[1]
        U1(fai / 2) & targs[1]
        Rx(-np.pi / 2) & targs[1]
        H & targs[1]
        Rzz(np.pi / 2) & targs
        H & targs[1]
        H & targs[0]
        Ry(-np.pi / 2) & targs[0]
        Rz(-fai / 2) & targs[0]
    return gates



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/2/16 4:25 下午
# @Author  : Han Yu
# @File    : cnot_force_depth.py

import os

import numpy as np

from QuICT.qcda.optimization.cnot_local_force.utility.utility import apply_cx, count, generate_layer


def generate_json(n):
    """ find the best circuit by bfs

    Args:
        n(int): input qubits

    Returns:
        Circuit: optimal circuit

    """
    now = 0
    for i in range(n):
        now ^= (1 << (i * n + i))
    vis = np.zeros(1 << (n * n), dtype=int)
    vis[now] = 1
    pre = [None] * (1 << (n * n))
    out = {}
    queue = [now]
    out[now] = []
    ans = count(n)
    l = 0
    total = 1
    layers = generate_layer(n)
    while True:
        now = queue[l]
        for i in range(len(layers)):
            layer = layers[i]
            for cx in layer:
                new_state = apply_cx(now, cx[0], cx[1], n)
            if vis[new_state] == 0:
                pre[new_state] = (now, i)
                vis[new_state] = vis[now] + 1
                queue.append(new_state)
                paths = []
                total += 1
                while new_state != queue[0]:
                    paths.extend(layers[pre[new_state][1]])
                    new_state = pre[new_state][0]
                out[queue[-1]] = []
                for index in range(len(paths) - 1, -1, -1):
                    out[queue[-1]].append((paths[index]))

        if total == ans:
            break
        l += 1

    keys = out.keys()
    with open(f"{os.path.dirname(os.path.abspath(__file__))}{os.path.sep}{n}qubit_cnot_depth.inf", 'w') as file:
        file.write(";")
        for key in keys:
            string = f"{key}:"
            tuples = out[key]
            len_tuples = len(tuples)
            for i in range(len_tuples):
                _tuple = tuples[i]
                string += f"{_tuple[0] * 5 + _tuple[1]}"
                if i + 1 < len_tuples:
                    string += ','
                else:
                    string += ';'
            if len_tuples == 0:
                string += ';'
            file.write(string)



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/3/30 11:13
# @Author  : Han Yu
# @File    : _IBMQStyle.py

from copy import copy
from warnings import warn


class DefaultStyle:
    """IBM Design Style colors
    """

    def __init__(self):
        # Set colors
        basis_color = '#FA74A6'
        clifford_color = '#6FA4FF'
        non_gate_color = '#000000'
        other_color = '#BB8BFF'
        pauli_color = '#05BAB6'
        iden_color = '#05BAB6'

        self.name = 'iqx'
        self.tc = '#000000'
        self.sc = '#000000'
        self.lc = '#000000'
        self.not_gate_lc = '#ffffff'
        self.cc = '#778899'
        self.gc = other_color
        self.gt = '#000000'
        self.bc = '#bdbdbd'
        self.bg = '#ffffff'
        self.edge_color = None
        self.math_fs = 15
        self.fs = 13
        self.sfs = 8
        self.colored_add_width = 0.2
        self.disptex = {
            'id': 'Id',
            'u0': 'U_0',
            'u1': 'U_1',
            'u2': 'U_2',
            'u3': 'U_3',
            'x': 'X',
            'y': 'Y',
            'z': 'Z',
            'h': 'H',
            's': 'S',
            'sdg': 'S^\\dagger',
            't': 'T',
            'tdg': 'T^\\dagger',
            'rx': 'R_x',
            'ry': 'R_y',
            'rz': 'R_z',
            'reset': '\\left|0\\right\\rangle',
            'barrier': 'barrier',
            'Permutation gate': 'perm',
            'Unitary gate': 'unitary',
            'phase': 'phase',
            'Custom gate': 'custom'
        }
        self.dispcol = {
            'u0': basis_color,
            'u1': basis_color,
            'u2': basis_color,
            'u3': basis_color,
            'id': iden_color,
            'x': pauli_color,
            'y': pauli_color,
            'z': pauli_color,
            'h': clifford_color,
            'cx': clifford_color,
            's': clifford_color,
            'sdg': clifford_color,
            't': other_color,
            'tdg': other_color,
            'rx': other_color,
            'ry': other_color,
            'rz': other_color,
            'reset': non_gate_color,
            'barrier': other_color,
            'target': '#ffffff',
            'swap': other_color,
            'multi': other_color,
            'meas': non_gate_color,
            'Permutation gate': other_color,
            'Unitary gate': other_color,
            'phase': other_color,
            'Custom gate': other_color
        }
        self.latexmode = False
        self.fold = None  # To be removed after 0.10 is released
        self.bundle = True
        self.index = False
        self.figwidth = -1
        self.dpi = 150
        self.margin = [2.0, 0.1, 0.1, 0.3]
        self.cline = 'doublet'

    def set_style(self, style_dic):
        dic = copy(style_dic)
        self.tc = dic.pop('textcolor', self.tc)
        self.sc = dic.pop('subtextcolor', self.sc)
        self.lc = dic.pop('linecolor', self.lc)
        self.cc = dic.pop('creglinecolor', self.cc)
        self.gt = dic.pop('gatetextcolor', self.tc)
        self.gc = dic.pop('gatefacecolor', self.gc)
        self.bc = dic.pop('barrierfacecolor', self.bc)
        self.bg = dic.pop('backgroundcolor', self.bg)
        self.fs = dic.pop('fontsize', self.fs)
        self.sfs = dic.pop('subfontsize', self.sfs)
        self.disptex = dic.pop('displaytext', self.disptex)
        self.dispcol = dic.pop('displaycolor', self.dispcol)
        self.latexmode = dic.pop('latexdrawerstyle', self.latexmode)
        self.bundle = dic.pop('cregbundle', self.bundle)
        self.index = dic.pop('showindex', self.index)
        self.figwidth = dic.pop('figwidth', self.figwidth)
        self.dpi = dic.pop('dpi', self.dpi)
        self.margin = dic.pop('margin', self.margin)
        self.cline = dic.pop('creglinestyle', self.cline)
        if 'fold' in dic:
            warn('The key "fold" in the argument "style" is being replaced by the argument "fold"',
                 DeprecationWarning, 5)
            self.fold = dic.pop('fold', self.fold)
            if self.fold < 2:
                self.fold = -1

        if dic:
            warn('style option/s ({}) is/are not supported'.format(', '.join(dic.keys())),
                 DeprecationWarning, 2)



from typing import Tuple

from QuICT.core import *
from QuICT.core.gate import CompositeGate
from QuICT.core.gate.gate import BasicGate

from ..common import LayoutInfo
from ..common.circuit_info import CircuitInfo
from .mcts_tree_node import MCTSTreeNode


class MCTSTree:
    def __init__(
        self,
        circuit_info: CircuitInfo,
        layout_info: LayoutInfo,
        bp_num: int,
        sim_cnt: int,
        sim_gate_num: int,
        gamma: float,
        epsilon: float,
        c: float,
    ) -> None:
        self._root = MCTSTreeNode(
            circuit_info=circuit_info,
            layout_info=layout_info,
        )
        self._layout_info = layout_info

        self._bp_num = bp_num
        self._sim_cnt = sim_cnt
        self._sim_gate_num = sim_gate_num
        self._c = c
        self._gamma = gamma
        self._epsilon = epsilon

    def step(self) -> Tuple[Tuple[int, int], CompositeGate, bool]:
        """Execute multiple search rounds.
        Return a swap selection and update current search tree root.

        Returns:
            Tuple[Tuple[int, int], bool]: Selected swap action, whether mapping ends.
        """
        for _ in range(self._bp_num):
            cur = self._root
            while len(cur.children) > 0:
                cur = cur.select_child(self._c)
            if cur.is_terminated_node():
                break
            cur.expand()
            for leaf in cur.children.values():
                leaf.simulate(
                    sim_cnt=self._sim_cnt,
                    sim_gate_num=self._sim_gate_num,
                    gamma=self._gamma,
                    epsilon=self._epsilon,
                )
                leaf.back_propagate(self._gamma)
        action, self._root, terminated = self._root.best_child()
        assert self._layout_info.topo_graph.has_edge(*action), "Illegal action!"
        for gate in self._root.executed_circ.gates:
            gate: BasicGate
            if gate.controls + gate.targets == 1:
                continue
            a, b = gate.cargs + gate.targs
            assert self._layout_info.topo_graph.has_edge(
                a, b
            ), "Wrong executed circuit!"
        return action, self._root.executed_circ, terminated



from collections import defaultdict
from typing import Dict, List, Union

import numpy as np


class Graph:
    """ The graph descript position space and action space for quantum random walk """
    @property
    def position(self) -> int:
        return self._nodes

    @property
    def position_qubits(self) -> int:
        return int(np.ceil(np.log2(self._nodes)))

    @property
    def action_space(self) -> int:
        return len(self._edges[0])

    @property
    def action_qubits(self) -> int:
        if self._operators is None:
            return int(np.ceil(np.log2(self.action_space)))
        else:
            shape = self._operators[0][0].shape
            return int(np.ceil(np.log2(shape[0])))

    @property
    def edges(self) -> dict:
        return self._edges

    @property
    def operators(self) -> dict:
        return self._operators

    @property
    def switched_time(self) -> int:
        return self._switched_time

    def __init__(
        self,
        position: int,
        edges: Union[List, Dict] = None,
        operators: Union[List, Dict] = None,
        switched_time: int = -1
    ):
        """ Initial the random walk graph.

        Args:
            position (int): The number of graph's nodes.
            edges (Union[List, Dict], optional): The edges of each node. Defaults to None.
            operators (Union[List, Dict], optional): The operators of each node. Defaults to None.
            switched_time (int, optional): The number of steps of each coin operator in the nodes.
                Defaults to -1, means not switch coin operator.
        """
        self._nodes = position
        self._switched_time = switched_time
        self._edges = defaultdict(list)
        if edges is not None:
            iterator = enumerate(edges) if isinstance(edges, list) else edges.items()
            for idx, edge in iterator:
                assert isinstance(edge, list)
                self._edges[idx] = edge

        self._operators = defaultdict(list) if operators is not None else operators
        if operators is not None:
            iterator = enumerate(operators) if isinstance(operators, list) else operators.items()
            assert len(iterator) == self._nodes, "The number of operators should equal to position"
            for idx, operator in iterator:
                assert self.operator_validation(operator), "The operator should be 1 or more unitary matrix."
                self._operators[idx] = operator if isinstance(operator, list) else [operator]

    def __str__(self):
        return str(self._edges) + "\nOperators: " + str(self._operators)

    def operator_validation(self, operator: Union[List, np.ndarray]) -> bool:
        """ Validate the operators """
        if isinstance(operator, np.ndarray):
            return self._operator_validation(operator)

        for op in operator:
            if not self._operator_validation(op):
                return False

        return True

    def _operator_validation(self, operator: np.ndarray):
        shape = operator.shape
        log2_shape = int(np.ceil(np.log2(shape[0])))

        return (
            shape[0] == shape[1] and
            shape[0] == (1 << log2_shape) and
            np.allclose(np.eye(shape[0]), operator.dot(operator.T.conj()))
        )

    def add_operator(self, u: int, operator: Union[List, np.ndarray]):
        """ Add operator to a vector. """
        assert u >= 0 and u <= self._nodes
        assert self.operator_validation(operator), "The operator should be 1 or more unitary matrix."
        if isinstance(operator, np.ndarray):
            operator = [operator]

        for op in operator:
            self._operators[u].append(op)

    def add_edge(self, u: int, v: int):
        """ Add edge """
        assert u >= 0 and u <= self._nodes
        assert v >= 0 and v <= self._nodes

        if v not in self._edges[u]:
            self._edges[u].append(v)

    def del_edge(self, u: int, v: int):
        """ Remove edge """
        assert u >= 0 and u <= self._nodes
        assert v >= 0 and v <= self._nodes

        if v in self._edges[u]:
            self._edges[u].remove(v)

    def validation(self) -> bool:
        """ Validate that all vectors has same number of edges. """
        edge_size = len(self._edges[0])
        for edge in self._edges.values():
            if len(edge) != edge_size:
                return False

        return True



from collections import Iterable
from functools import cached_property
from itertools import chain
from typing import List

from QuICT.core.utils import GateType
from QuICT.core import Circuit

from .matching_dag_circuit import Match, MatchingDAGCircuit, MatchingDAGNode


class CircuitCostMeasure(object):
    """
    Measure of gate cost in quantum circuit.
    """

    NISQ_GATE_COST = {
        GateType.id: 0, GateType.x: 1, GateType.y: 1, GateType.z: 1, GateType.h: 1,
        GateType.t: 1, GateType.tdg: 1, GateType.s: 1, GateType.sdg: 1, GateType.u1: 1,
        GateType.u2: 2, GateType.u3: 2, GateType.rx: 1, GateType.ry: 1, GateType.rz: 1,
        GateType.cx: 2, GateType.cy: 4, GateType.cz: 4, GateType.ch: 8, GateType.swap: 6,
        GateType.cswap: 8, GateType.rxx: 9, GateType.ryy: 9, GateType.rzz: 5,
        GateType.cu3: 10, GateType.ccx: 21, GateType.measure: 9
    }

    # FIXME fill other gates
    FT_GATE_COST = {
        GateType.t: 1, GateType.tdg: 1
    }

    def __init__(self, cost_dict=None, target_device='nisq'):
        """
        Args:
            cost_dict(Dict[GateType, int]): Cost of each gate type. It a gate type is not specified,
                its cost is treated as 0.
            target_device(str): Type of target device. Support 'nisq' and 'fault_tolerant'. It will
                be ignored if `cost_dict` is given.
        """
        if cost_dict is not None:
            self.cost_dict = cost_dict
        elif target_device == 'nisq':
            self.cost_dict = self.NISQ_GATE_COST
        elif target_device == 'fault_tolerant':
            self.cost_dict = self.FT_GATE_COST
        else:
            assert f'target device {target_device} not supported. It must be nisq or fault_tolerant'

    def __getitem__(self, gate_type):
        """
        Get cost of a gate type. Subscript can be a GateType or string of a gate type.

        Args:
            gate_type(GateType/str): Gate type

        Returns:
            int: Cost of the gate type
        """
        if isinstance(gate_type, str):
            gate_type = GateType(gate_type)

        if gate_type in self.cost_dict:
            return self.cost_dict[gate_type]
        else:
            return 0

    def cost(self, circuit):
        """
        Calculate the cost of a circuit.

        Args:
            circuit(DAGCircuit/Circuit/CompositeGate/Iterable[BasicGate]/Iterable[DAGCircuitNode]): the circuit

        Returns:
            int: Cost of the circuit
        """

        if isinstance(circuit, Iterable):
            return sum(self[n.gate.type] for n in circuit)
        else:
            return sum(self[n.gate.type] for n in circuit.gates)


class Substitution:
    """
    Class for a substitution.
    """
    def __init__(self,
                 circuit: MatchingDAGCircuit,
                 template: MatchingDAGCircuit,
                 match: Match,
                 measure: CircuitCostMeasure):
        """
        Args:
            circuit(MatchingDAGCircuit): the circuit to match
            template(MatchingDAGCircuit): the template to be matched
            match(Match): the matching for this substitution
            measure(CircuitCostMeasure): cost measure
        """
        self.circuit = circuit
        self.template = template
        self.match = match
        self.measure = measure

    @cached_property
    def cost(self):
        """
        Returns:
            int: Cost reduced if when substitution is done.
        """
        old_nodes = self.match.template_nodes
        new_nodes = set(range(self.template.size)) - old_nodes
        old_cost = sum(self.measure[self.template.get_node(n).type] for n in old_nodes)
        new_cost = sum(self.measure[self.template.get_node(n).type] for n in new_nodes)
        return old_cost - new_cost

    @cached_property
    def pred(self):
        """
        Returns:
            List[int]: predecessors (direct and indirect) of matched nodes in the circuit
        """
        return sorted(self.pred_set)

    @cached_property
    def pred_set(self):
        """
        Returns:
            Set[int]: predecessors (direct and indirect) of matched nodes in the circuit
        """

        return self.circuit.all_predecessors(self.match.circuit_nodes)

    def cmp_with(self, other):
        """
        Whether `self` comes before `other`. Namely, whether any matched node of
        `self` is a predecessor of matched nodes of `other`.

        Args:
            other(Substitution): the other substitution to compare

        Returns:
            int: -1 if `self` comes before `other`. Otherwise 1.
        """

        assert isinstance(other, Substitution)
        return -1 if self.match.circuit_nodes & other.pred_set else 1

    def get_substitution(self):
        """
        Get the substitution of circuit for the matching.
        """

        pred = self.template.all_predecessors(self.match.template_nodes)
        succ = set(self.template.nodes()) - self.match.template_nodes - pred

        circ = Circuit(self.circuit.width)
        for node_id in chain(sorted(pred, reverse=True), sorted(succ, reverse=True)):
            # inverse every gates not matched
            node: MatchingDAGNode = self.template.get_node(node_id)
            qubits = list([self.match.qubit_mapping[i] for i in node.qargs])
            node.gate.inverse().copy() | circ(qubits)
        return circ


class TemplateSubstitution:

    @classmethod
    def _sort_sub_list(cls, sub_list: List[Substitution]):
        """
        Sort all substitution in topological order.
        """
        res = []
        while len(sub_list) > 0:
            # find a sub that is not preceded by any other sub
            found = -1
            for idx, cur in enumerate(sub_list):
                if cur is not None and all(s.cmp_with(cur) > 0 for s in sub_list):
                    found = idx
                    break

            # add the found sub to the resulting list
            res.append(sub_list[found])
            sub_list = sub_list[:found] + sub_list[found + 1:]

        return res

    @classmethod
    def _calc_sub_list(cls,
                       circuit: MatchingDAGCircuit,
                       template: MatchingDAGCircuit,
                       match_list: List[Match],
                       measure: CircuitCostMeasure):
        """
        Greedily calculate a maximal and compatible sub list for found matches.
        """

        all_subs = [Substitution(circuit, template, m, measure) for m in match_list]

        sub_list = []
        visited_nodes = set()
        # enumerate subs that have positive `cost` in decreasing order
        for sub in sorted(filter(lambda s: s.cost > 0, all_subs),
                          key=lambda s: s.cost,
                          reverse=True):
            cur_nodes = sub.match.circuit_nodes

            # subs must not intersect
            if visited_nodes & cur_nodes:
                continue

            # if two subs both have some node precede the other, it is not valid
            flag_intersect = False
            for other_sub in sub_list:
                if (other_sub.pred_set & cur_nodes) and \
                        (sub.pred_set & other_sub.match.circuit_nodes):
                    flag_intersect = True
                    break

            if not flag_intersect:
                visited_nodes |= cur_nodes
                sub_list.append(sub)

        return cls._sort_sub_list(sub_list)

    @classmethod
    def execute(cls,
                circuit: MatchingDAGCircuit,
                template: MatchingDAGCircuit,
                match_list: List[Match],
                measure: CircuitCostMeasure):

        sub_list = cls._calc_sub_list(circuit, template, match_list, measure)

        new_circ = Circuit(circuit.width)
        visited_nodes = set()

        # Given a sorted valid sub list, every time append unvisited predecessors to the
        # resulting circuit and then append the inverse circuit of the sub.
        for sub in sub_list:
            for node_id in filter(lambda x: x not in visited_nodes, sub.pred):
                node: MatchingDAGNode = circuit.get_node(node_id)
                new_circ.append(node.gate.copy())

            new_circ.extend(sub.get_substitution().gates)

            visited_nodes.update(sub.pred)
            visited_nodes.update(sub.match.circuit_nodes)

        for node_id in filter(lambda x: x not in visited_nodes, circuit.nodes()):
            node: MatchingDAGNode = circuit.get_node(node_id)
            new_circ.append(node.gate.copy())

        return MatchingDAGCircuit(new_circ)



import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import CompositeGate, GateType, GPhase, X, CX, Ry, Rz
from QuICT.core.gate.backend import MCTWithoutAux, UniformlyRotation
from QuICT.qcda.synthesis.unitary_decomposition import UnitaryDecomposition
from QuICT.simulation.state_vector import ConstantStateVectorSimulator


class QuantumStatePreparation(object):
    """
    For a given quantum state |psi>, create a CompositeGate C that |psi> = C |0>
    """
    def __init__(self, method='unitary_decomposition'):
        """
        Choose the method between reference [1] and [2], designing circuit of
        quantum state preparation with uniformly gates and unitary decomposition respectively

        Args:
            method(str, optional): chosen method in ['uniformly_gates', 'unitary_decomposition']

        Reference:
            [1] https://arxiv.org/abs/quant-ph/0407010v1
            [2] https://arxiv.org/abs/1003.5760
        """
        assert method in ['uniformly_gates', 'unitary_decomposition'],\
            ValueError('Invalid quantum state preparation method')
        self.method = method

    def execute(self, state_vector):
        """
        Args:
            state_vector(np.ndarray): the statevector to be prapared

        Returns:
            CompositeGate: the preparation CompositeGate
        """
        if self.method == 'uniformly_gates':
            return self._with_uniformly_gates(state_vector)
        if self.method == 'unitary_decomposition':
            return self._with_unitary_decomposition(state_vector)

    def _with_uniformly_gates(self, state_vector):
        """
        Quantum state preparation with uniformly gates

        Args:
            state_vector(np.ndarray): the statevector to be prapared

        Returns:
            CompositeGate: the preparation CompositeGate

        Reference:
            https://arxiv.org/abs/quant-ph/0407010v1
        """
        state_vector = np.array(state_vector)
        num_qubits = int(round(np.log2(state_vector.size)))
        assert state_vector.ndim == 1 and 1 << num_qubits == state_vector.size,\
            ValueError('Quantum state should be an array with length 2^n')

        gates = CompositeGate()
        omega = np.angle(state_vector)
        state_vector = np.abs(state_vector)
        # Now for the non-negative real state_vector
        URy = UniformlyRotation(GateType.ry)
        denominator = np.linalg.norm(state_vector)
        for k in range(num_qubits - 1, -1, -1):
            numerator = np.linalg.norm(state_vector.reshape(1 << num_qubits - k, 1 << k), axis=1)
            alpha = np.where(np.isclose(denominator, 0), 0, 2 * np.arcsin(numerator[1::2] / denominator))
            gates.extend(URy.execute(alpha))
            denominator = numerator
        # If state_vector is real and non-negative, no UniformlyRz will be needed.
        URz = UniformlyRotation(GateType.rz)
        if not np.allclose(omega, 0):
            for k in range(num_qubits):
                alpha = np.sum(omega.reshape(1 << num_qubits - k, 1 << k), axis=1)
                alpha = (alpha[1::2] - alpha[0::2]) / (1 << k)
                gates.extend(URz.execute(alpha))
            gates.append(GPhase(np.average(omega)) & 0)

        return gates

    def _with_unitary_decomposition(self, state_vector):
        """
        Quantum state preparation with unitary decomposition

        Args:
            state_vector(np.ndarray): the statevector to be prapared

        Returns:
            CompositeGate: the preparation CompositeGate

        Reference:
            https://arxiv.org/abs/1003.5760
        """
        state_vector = np.array(state_vector)
        num_qubits = int(round(np.log2(state_vector.size)))
        assert state_vector.ndim == 1 and 1 << num_qubits == state_vector.size,\
            ValueError('Quantum state should be an array with length 2^n')

        first_half = num_qubits // 2 if np.mod(num_qubits, 2) == 0 else (num_qubits - 1) // 2
        last_half = num_qubits - first_half
        state_vector = state_vector.reshape(1 << first_half, 1 << last_half)
        # Schmidt decomposition
        U, d, V = np.linalg.svd(state_vector)

        gates = CompositeGate()
        # Phase 1
        gates.extend(self._with_uniformly_gates(d))
        # Phase 2
        with gates:
            for i in range(first_half):
                CX & [i, i + first_half]
        UD = UnitaryDecomposition()
        # Phase 3
        U_gates, _ = UD.execute(U)
        gates.extend(U_gates)
        # Phase 4
        if np.mod(num_qubits, 2) != 0:
            V = V[np.arange(1 << last_half).reshape(2, 1 << last_half - 1).T.flatten()]
        V_gates, _ = UD.execute(V.T)
        V_gates & list(range(first_half, num_qubits))
        gates.extend(V_gates)

        return gates


class SparseQuantumStatePreparation(object):
    """
    For a sparse quantum state |psi>, i.e. only a few of entries in the statevector is not zero,
    create a circuit C that |psi> = C |0>
    """
    def __init__(self, input_format='state_vector'):
        """
        Despite the original state vector, we could also use a 'state array' to describe a sparse quantum state.
        That is, each element [basis, amplitude] in the state array gives the amplitude on the computational basis,
        where the bases are boolean strings and not shown amplitudes are zeros.
        Be careful that no duplicated basis is allowed.

        Args:
            input_mode(str, optional): chosen input in ['state_array', 'state_vector']

        Reference:
            https://ieeexplore.ieee.org/abstract/document/9586240/metrics
        """
        assert input_format in ['state_array', 'state_vector'],\
            ValueError('Invalid sparse quantum state preparation input format')
        self.input_format = input_format

    def execute(self, state_array) -> CompositeGate:
        """
        Quantum state preparation for sparse state

        Args:
            state_array: input quantum state, whose format is decided by the input_format

        Returns:
            CompositeGate: the preparation CompositeGate
        """
        if self.input_format == 'state_array':
            # By the dict rule, only the last value of duplicated key works.
            state = dict(state_array)
            width = len(list(state.keys())[0])
            for basis in state.keys():
                assert len(basis) == width, ValueError('Bases must have the same length.')
            state_vector = self.dict_to_statevector(state, width)
        if self.input_format == 'state_vector':
            state, width = self.statevector_to_dict(state_array)
            state_vector = np.array(state_array)

        gates = CompositeGate()
        simulator = ConstantStateVectorSimulator()
        while len(state) > 1:
            gates_last = self.reduce_state(state, width)
            state_vector = self.dict_to_statevector(state, width)
            cir = Circuit(width)
            cir.extend(gates_last)
            state_vector = simulator.run(cir, state_vector).get()
            state, width = self.statevector_to_dict(state_vector)
            gates.extend(gates_last)

        x = list(state.keys())[0]
        for b in range(width):
            if x[b] == '1':
                X & b | gates
        phase = -np.angle(state[x])
        GPhase(phase) & 0 | gates
        return gates.inverse()

    def reduce_state(self, state: dict, width: int) -> CompositeGate:
        """
        Algorithm 1 in the reference, to reduce the non-zero entries in the state vector.

        Args:
            state(dict): a sparse state dict
            width(int): the width of the state

        Returns:
            CompositeGate: a CompositeGate that reduce the state dict
        """
        gates = CompositeGate()
        self.dif_qubits = []
        self.dif_values = []

        x1 = self.dif_qubits_values(state.keys(), width)
        dif = self.dif_qubits.pop()
        self.dif_values.pop()

        T_prime = []
        for x in state.keys():
            tp = True
            for b, v in zip(self.dif_qubits, self.dif_values):
                if x[b] != str(v):
                    tp = False
                    break
            if tp:
                T_prime.append(x)
        T_prime.remove(x1)
        x2 = self.dif_qubits_values(T_prime, width)

        if x1[dif] != '1':
            X & dif | gates
        for b in range(width):
            if b == dif:
                continue
            if x1[b] != x2[b]:
                CX & [dif, b] | gates
        for b in self.dif_qubits:
            if x2[b] != '1':
                X & b | gates

        mcg, _ = self.multicontrol_G(len(self.dif_qubits), state[x2], state[x1])
        mcg & (self.dif_qubits + [dif])
        mcg | gates
        return gates

    def dif_qubits_values(self, basis_list: list, width: int):
        while len(basis_list) > 1:
            dif_b = None
            dif_T0 = []
            dif_T1 = []
            for b in range(width):
                T0 = []
                T1 = []
                for x in basis_list:
                    if x[b] == '0':
                        T0.append(x)
                    else:
                        T1.append(x)
                if T0 and T1 and np.abs(len(dif_T0) - len(dif_T1)) <= np.abs(len(T0) - len(T1)):
                    dif_b = b
                    dif_T0 = T0
                    dif_T1 = T1
            self.dif_qubits.append(dif_b)
            if len(dif_T0) <= len(dif_T1):
                self.dif_values.append(0)
                basis_list = dif_T0
            else:
                self.dif_values.append(1)
                basis_list = dif_T1
        return basis_list[0]

    @staticmethod
    def multicontrol_G(control: int, alpha: complex, beta: complex) -> CompositeGate:
        """
        Create a CompositeGate that maps alpha|0> + beta|1> to e^{i phase} |0> with some control qubits

        Args:
            control(int): the number of control qubits
            alpha(complex): alpha in the initial state
            beta(complex): beta in the initial stats

        Returns:
            CompositeGate: the required CompositeGate

        References:
            https://arxiv.org/abs/quant-ph/9503016
        """
        arr = np.array([alpha, beta])
        arr /= np.linalg.norm(arr)
        alpha, beta = arr
        omega = 2 * np.arcsin(np.abs(alpha))
        gamma = np.angle(alpha) - np.angle(beta)

        gates = CompositeGate()
        Rz(gamma) & control | gates
        Ry(omega / 2) & control | gates
        MCTWithoutAux().execute(control + 1) | gates
        Ry(-omega / 2) & control | gates
        Rz(-gamma) & control | gates

        return gates, np.angle(alpha)

    @staticmethod
    def statevector_to_dict(state_vector):
        state = dict()
        width = int(round(np.log2(state_vector.size)))
        for basis, amp in enumerate(state_vector):
            if not np.isclose(amp, 0):
                state[bin(basis)[2:].zfill(width)] = amp
        return state, width

    @staticmethod
    def dict_to_statevector(state: dict, width):
        state_vector = np.zeros(1 << width, dtype=complex)
        for basis, amp in state.items():
            if not np.isclose(amp, 0):
                basis = int(basis, base=2)
                state_vector[basis] = amp
        return state_vector



from typing import List, Tuple

from .matching_dag_circuit import MatchingDAGCircuit, MatchingDAGNode


class ForwardMatch:
    """
    Class for forward matching.
    """

    @classmethod
    def _insert_match_nodes(cls, match_nodes: List[MatchingDAGNode], node: MatchingDAGNode):
        """
        Insert a node to `matched_node` list. Then sorted the list in increasing order of
        the first element of `successors_to_visit`.
        """
        if len(node.successors_to_visit) > 0:
            match_nodes.append(node)
            match_nodes.sort(key=lambda n: n.successors_to_visit[0])

    @classmethod
    def _find_candidates(cls,
                         match: List[Tuple[int, int]],
                         template: MatchingDAGCircuit,
                         t_node_id: int) -> List[MatchingDAGNode]:
        """
        Find valid candidates for next match
        """
        match = {match[i][0] for i in range(len(match))}

        block = set()
        for u in match - {t_node_id}:
            for v in template.get_node(u).successors:
                if v not in match:
                    block |= template.all_successors(v)

        id_set = set(template.get_node(t_node_id).successors) - match - block
        return [template.get_node(i) for i in id_set]

    @classmethod
    def execute(cls,
                circuit: MatchingDAGCircuit,
                template: MatchingDAGCircuit,
                c_node_id: int,
                t_node_id: int,
                qubit_mapping: List[int]):
        """
        Execute Forward matching.

        Args:
            circuit(MatchingDAGCircuit): the circuit to match
            template(MatchingDAGCircuit): the template to be matched
            c_node_id(int): the starting node of `circuit`
            t_node_id(int): the starting node of `template`
            qubit_mapping(List[int]): qubit mapping
                (qubit i in `temlate` is mapped to qubit_mapping[i] is circuit)

        Returns:
            List[Tuple[int, int]]: the maximal match (template node id, circuit node id)'s
        """

        # initialize properties
        circuit.init_forward_matching(c_node_id, t_node_id, s2v_enabled=True)
        template.init_forward_matching(t_node_id, c_node_id)

        match = [(t_node_id, c_node_id)]
        match_nodes = [circuit.get_node(c_node_id)]
        while len(match_nodes) > 0:
            # every time match the first element of successors_to_visit of one node
            cur_node: MatchingDAGNode = match_nodes[0]
            match_nodes.pop(0)

            nxt_node_id = cur_node.pop_successors_to_visit()
            if nxt_node_id is None:
                continue
            nxt_node = circuit.get_node(nxt_node_id)

            if cur_node.successors_to_visit:
                cls._insert_match_nodes(match_nodes, cur_node)

            if not nxt_node.matchable():
                continue

            cands: List[MatchingDAGNode] = cls._find_candidates(match, template, cur_node.matched_with)
            success = False
            for t_nxt_node in cands:
                if t_nxt_node.compare_with(nxt_node, qubit_mapping):
                    t_nxt_node.matched_with = nxt_node.id
                    nxt_node.matched_with = t_nxt_node.id
                    match.append((t_nxt_node.id, nxt_node.id))

                    # calc nxt_node.successors_to_visit
                    nxt_node.successors_to_visit = sorted(list(filter(
                        lambda x: circuit.get_node(x).matchable,
                        nxt_node.successors
                    )))
                    # match_nodes.append(nxt_node)
                    cls._insert_match_nodes(match_nodes, nxt_node)

                    success = True
                    break

            if not success:
                # if not succeeded, block the node and all its successors
                nxt_node.is_blocked = True
                for node_id in circuit.all_successors(nxt_node.id):
                    succ_node = circuit.get_node(node_id)

                    succ_node.is_blocked = True
                    # remove matched info
                    if succ_node.matched_with:
                        t = succ_node.matched_with
                        template.get_node(t).matched_with = None
                        succ_node.matched_with = None
                        match.remove((t, succ_node.id))

        return match



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/12/27 10:45 下午
# @Author  : Han Yu
# @File    : uniformRotation.py

from typing import *
import numpy as np

from QuICT.core.gate import GateType, build_gate, CompositeGate, CX


class UniformlyRotation(object):
    """
    Implements the uniformly Ry or Rz gate

    Reference:
        https://arxiv.org/abs/quant-ph/0504100 Fig4 a)
    """
    def __init__(self, gate_type=None):
        """
        Args:
            gate_type(GateType): the type of uniformly gate, Ry or Rz
        """
        assert gate_type in [GateType.ry, GateType.rz], ValueError('Invalid gate_type')
        self.gate_type = gate_type

    def execute(self, angle_list):
        """
        Args:
            angle_list(list<float>): the angles of Ry or Rz gates

        Returns:
            CompositeGate: CompositeGate that implements the uniformly gate
        """
        angle_list = list(angle_list)
        n = int(np.log2(len(angle_list))) + 1
        if 1 << (n - 1) != len(angle_list):
            raise Exception("the number of parameters unmatched.")
        return self.uniformly_rotation(0, n, angle_list, self.gate_type)

    def uniformly_rotation(
        self,
        low: int,
        high: int,
        angles: List[float],
        gate_type: int,
        is_left_cnot: bool = False
    ) -> CompositeGate:
        """
        synthesis uniformlyRotation gate, bits range [low, high)

        Args:
            low(int): the left range low
            high(int): the right range high
            angles(list<float>): the list of angle y
            gate_type(int): the gateType (Rz or Ry)
            is_left_cnot(bool): is cnot left decomposition
        Returns:
            gateSet: the synthesis gate list
        """
        return self.inner_uniformly_rotation(low, high, angles, gate_type, True, is_left_cnot)

    def inner_uniformly_rotation(
        self,
        low: int,
        high: int,
        angles: List[float],
        gate_type: int,
        is_first_level: bool,
        is_left_cnot: bool = False
    ) -> CompositeGate:
        if low + 1 == high:
            rot = build_gate(gate_type, low, [angles[0].real])
            gates = CompositeGate()
            gates.append(rot)
            return gates
        length = len(angles) // 2
        Rxp = []
        Rxn = []
        for i in range(length):
            Rxp.append((angles[i] + angles[i + length]) / 2)
            Rxn.append((angles[i] - angles[i + length]) / 2)
        if is_first_level:
            if is_left_cnot:
                gates = CompositeGate()
                CX & [low, high - 1] | gates
                gates.extend(self.inner_uniformly_rotation(low + 1, high, Rxn, gate_type, False, False))
                CX & [low, high - 1] | gates
                gates.extend(self.inner_uniformly_rotation(low + 1, high, Rxp, gate_type, False, True))
            else:
                gates = self.inner_uniformly_rotation(low + 1, high, Rxp, gate_type, False, False)
                CX & [low, high - 1] | gates
                gates.extend(self.inner_uniformly_rotation(low + 1, high, Rxn, gate_type, False, True))
                CX & [low, high - 1] | gates
        elif is_left_cnot:
            gates = self.inner_uniformly_rotation(low + 1, high, Rxn, gate_type, False, False)
            CX & [low, high - 1] | gates
            gates.extend(self.inner_uniformly_rotation(low + 1, high, Rxp, gate_type, False, True))
        else:
            gates = self.inner_uniformly_rotation(low + 1, high, Rxp, gate_type, False, False)
            CX & [low, high - 1] | gates
            gates.extend(self.inner_uniformly_rotation(low + 1, high, Rxn, gate_type, False, True))
        return gates



import numpy as np

from QuICT.core.gate import CompositeGate, H, X, CX, CU1


class MCTWithoutAux(object):
    """
    Decomposition of n-qubit Toffoli gates with no ancillary qubit

    Reference:
        https://arxiv.org/abs/1303.3557
    """
    def execute(self, n):
        """
        Args:
            n(int): the number of qubits, be aware that n means (n-1)-control NOT

        Returns:
            CompositeGate: the n-qubit Toffoli gate
        """
        gates = CompositeGate()
        if n == 1:
            X & 0 | gates
            return gates

        if n == 2:
            CX & [0, 1] | gates
            return gates

        for control in range(n - 2, 0, -1):
            for target in range(control + 1, n):
                if target == control + 1:
                    H & target | gates
                angle = np.pi / (2 ** (target - control))
                CU1(angle) & [control, target] | gates

        H & 1 | gates
        for target in range(1, n):
            angle = np.pi / (2 ** (target - 1))
            CU1(angle) & [0, target] | gates
        H & 1 | gates

        for control in range(1, n - 1, 1):
            for target in range(control + 1, n):
                angle = -np.pi / (2 ** (target - control))
                CU1(angle) & [control, target] | gates
                if target == control + 1:
                    H & target | gates

        for control in range(n - 3, 0, -1):
            for target in range(control + 1, n - 1):
                if target == control + 1:
                    H & target | gates
                angle = np.pi / (2 ** (target - control))
                CU1(angle) & [control, target] | gates

        H & 1 | gates
        for target in range(1, n - 1):
            angle = -np.pi / (2 ** (target - 1))
            CU1(angle) & [0, target] | gates
        H & 1 | gates

        for control in range(1, n - 2, 1):
            for target in range(control + 1, n - 1):
                angle = -np.pi / (2 ** (target - control))
                CU1(angle) & [control, target] | gates
                if target == control + 1:
                    H & target | gates

        return gates



#!/usr/bin/env python
# -*- coding:utf8 -*-

from typing import *
import numpy as np


class Merged:
    """
    Merged points in heap

    Attributes:
        deg(int) : degree of x
        nodes(List[int]) : node indices in a bipartite
    """

    def __init__(self, deg: int, nodes: List[int]):
        """
        Get a merged point

        Args:
            deg(int) : degree of x
            nodes(List[int]) : node indices in a bipartite
        """
        self.deg = deg
        self.nodes = nodes

    def __add__(self, other: "Merged") -> "Merged":
        """
        combine two merged points to a larger merged

        Args:
            other(Merged): Another Merged node.
        """
        d = self.deg + other.deg
        x = self.nodes + other.nodes
        return Merged(d, x)

    def __lt__(self, other):
        return self.deg < other.deg


def f2_matmul(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """
    Multiply 2 matrices in F_2.

    Args:
        a (np.ndarray): Input boolean matrix a.
        b (np.ndarray): Input boolean matrix b.

    Returns:
        np.ndarray: Production of ab over F_2.
    """
    a_ = np.array(a, dtype=int)
    b_ = np.array(b, dtype=int)
    c_ = a_ @ b_
    c_ %= 2
    c = np.array(c_, dtype=bool)
    return c


def f2_half_gaussian_elimination(mat_: np.ndarray) -> np.ndarray:
    """
    Gaussian elimination to convert a matrix into triangular form over F_2.
    Args:
        mat_ (np.ndarray): Input boolean matrix.

    Returns:
        np.ndarray: Eliminated triangular matrix.
    """
    mat: np.ndarray = mat_.copy()
    row_pivot = 0
    col_pivot = 0
    m = mat.shape[0]
    n = mat.shape[1]
    while row_pivot < m and col_pivot < n:
        if not mat[row_pivot, col_pivot]:
            for k in range(row_pivot + 1, m):
                if mat[k, col_pivot]:
                    mat[[k, row_pivot], :] = mat[[row_pivot, k], :]
                    break
        if mat[row_pivot, col_pivot]:
            for k in range(row_pivot + 1, m):
                if mat[k, col_pivot]:
                    mat[k, :] ^= mat[row_pivot, :]
            row_pivot += 1
            col_pivot += 1
        else:
            col_pivot += 1
    return mat


def f2_rank(mat_: np.ndarray) -> int:
    """
    Rank of a matrix over F_2.
    Args:
        mat_ (np.ndarray): Input boolean matrix.

    Returns:
        int: Rank of mat_ over F_2.
    """
    mat = f2_half_gaussian_elimination(mat_)
    rk = 0
    for i in range(mat.shape[0]):
        if np.any(mat[i, :]):
            rk += 1
    return rk


def f2_inverse(mat_: np.ndarray) -> np.ndarray:
    """
    Matrix inverse over F_2.
    Args:
        mat_ (np.ndarray): Input boolean matrix.

    Returns:
        np.ndarray: The inverse boolean matrix of mat_ over F_2.
    """
    n = mat_.shape[0]
    aug = np.empty(shape=(n, 2 * n), dtype=bool)
    aug[:, :n] = mat_
    aug[:, n:] = np.eye(n, dtype=bool)

    # gaussian elimination
    for i in range(n):
        if not aug[i, i]:
            for k in range(i + 1, n):
                if aug[k, i]:
                    aug[[k, i], :] = aug[[i, k], :]
                    break
        if not aug[i, i]:
            raise Exception("Matrix is not invertible!")
        for k in range(n):
            if i != k and aug[k, i]:
                aug[k, :] ^= aug[i, :]
    ret = aug[:, n:].copy()
    return ret



import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.core.operator import DataSwitch, DataSwitchType, DeviceTrigger, Multiply, SpecialGate


ALL_PENTY = 0.5
HALF_PENTY = 2
CARG_PENTY = 3
PROB_ADD = 1


class Transpile:
    """ Transpile the circuit by the number of devices.

    Args:
        ndev (int): The number of devices.
        cost_threshold (int, optional): The cost of exchanged splited qubits. Defaults to 3.
    """
    @property
    def ndev(self) -> int:
        """ Return the number of devices """
        return self._ndev

    @ndev.setter
    def ndev(self, ndev: int):
        self._ndev = ndev
        self._split_qb_num = int(np.log2(ndev))

    @property
    def cost_threshold(self) -> int:
        """ Return the cost of exchanged splited qubits. """
        return self._cost_threshold

    @cost_threshold.setter
    def cost_threshold(self, cost_threshold: int):
        self._cost_threshold = cost_threshold

    def __init__(self, ndev: int, cost_threshold: int = 3):
        self._ndev = ndev
        self._cost_threshold = cost_threshold
        self._split_qb_num = int(np.log2(ndev))
        assert 2 ** self._split_qb_num == self._ndev

    def _args_adjust(self, gate_args, outer_qubit_indexes: list, inner_qubit_indexes: list):
        """ Get the updated gate args """
        inner_arg, outer_arg = [], []
        for garg in gate_args:
            if garg in outer_qubit_indexes:
                outer_arg.append((outer_qubit_indexes.index(garg), garg))
            else:
                inner_arg.append(inner_qubit_indexes.index(garg))

        return inner_arg, outer_arg

    def _transpile(self, gates_by_depth: list, split_qubits: list, layer_intervals: list) -> Circuit:
        """ Generate the distributed circuit through the dynamic splited qubits.

        Args:
            gates_by_depth (list): The list of gates at same layers in circuit.
            split_qubits (list): The splited qubits' indexes for each layer intervals
            layer_intervals (list): The layer intervals, which have the same splited qubits

        Returns:
            Circuit: The distributed circuit.
        """
        inner_qubits = self._qubits - self._split_qb_num    # The qubits' number of the transpile circuit
        qubits_indexes = set(range(self._qubits))
        transpiled_circuit = Circuit(inner_qubits)
        start_point = 0
        for interval_idx, point in enumerate(layer_intervals):
            current_outer_qubits = split_qubits[interval_idx]   # The splited qubits' indexes in current interval
            if isinstance(current_outer_qubits, int):
                current_outer_qubits = [current_outer_qubits]

            # Currently inner qubits' indexes
            inner_qubit_indexes = list(qubits_indexes.difference(current_outer_qubits))
            for gates in gates_by_depth[start_point:point]:     # Deal with all gates in current layer interval
                for gate in gates:
                    if gate.type in [GateType.id, GateType.barrier]:
                        continue

                    gate_args = gate.cargs + gate.targs
                    inner_args, outer_arg = self._args_adjust(gate_args, current_outer_qubits, inner_qubit_indexes)
                    if len(outer_arg) == 0:
                        if gate.matrix_type == MatrixType.special:
                            SpecialGate(gate.type, inner_args) | transpiled_circuit
                        else:
                            gate = gate & inner_args
                            gate | transpiled_circuit

                        continue

                    # Deal with gate has qubits in splited qubits
                    if len(gate_args) == 1:
                        device_trigger = self._single_qubit_gate_transpile(gate, outer_arg[0])
                    elif len(gate_args) == 2:
                        device_trigger = self._double_qubits_gate_transpile(gate, inner_args, outer_arg)
                    else:
                        raise ValueError("Only supportted transpilt for gate less than 2 qubits.")

                    device_trigger | transpiled_circuit

            start_point = point
            # Deal with exchange inner qubits and outer qubits
            if interval_idx != len(layer_intervals) - 1 and split_qubits[interval_idx + 1] != current_outer_qubits[0]:
                updated_inner_args = inner_qubit_indexes.index(split_qubits[interval_idx + 1])
                self._split_qubit_exchange(1, inner_qubits - 1 - updated_inner_args) | transpiled_circuit
                perm_gate = self._permutation(current_outer_qubits[0], split_qubits[interval_idx + 1])
                perm_gate.build_gate() | transpiled_circuit

        return transpiled_circuit

    def _split_qubit_exchange(self, old_qubit_index: int, new_split_qubit: int) -> DeviceTrigger:
        """ Generate the DataSwitch about exchanged splited qubits for each device.

        Args:
            old_qubit_index (int): The index of splited qubit which has been replaced
            new_split_qubit (int): The index of new splited qubit.
        """
        dev_mapping = {}
        for index in range(self._ndev):
            splited_cgate = CompositeGate()
            dest = index ^ old_qubit_index
            _0_1 = int(index < dest)
            DataSwitch(dest, DataSwitchType.ctarg, {new_split_qubit: _0_1}) | splited_cgate

            dev_mapping[index] = splited_cgate

        return DeviceTrigger(dev_mapping)

    def _permutation(self, old, new) -> PermGate:
        """ Permutation State Vector after exchanged inner qubits and outer qubits """
        based_range = list(range(self._qubits))
        based_range.remove(old)
        based_range[based_range.index(new)] = old

        # re-order the based range by its indexes
        minimal_indexes = np.argsort(based_range)
        for idx, mini_idx in enumerate(minimal_indexes):
            based_range[mini_idx] = idx

        return Perm(self._qubits - self._split_qb_num, based_range)

    def _single_qubit_gate_transpile(
        self,
        gate: BasicGate,
        outer_qubits_indexes: tuple
    ) -> DeviceTrigger:
        """ Deal with single-qubits gate which in the outer qubit.

        Args:
            gate (BasicGate): The quantum gate.
            outer_qubits_indexes (tuple): The index
        """
        matrix_type = gate.matrix_type
        split_idx = 1 << (outer_qubits_indexes[0])

        dev_mapping = {}
        for index in range(self._ndev):
            _0_1 = index & split_idx
            destination = index ^ split_idx
            splited_cgate = CompositeGate()
            # Normal 1-qubit gates
            if matrix_type == MatrixType.normal:
                DataSwitch(destination, DataSwitchType.half) | splited_cgate
                gate | splited_cgate(0)
                DataSwitch(destination, DataSwitchType.half) | splited_cgate
            # diagonal 1-qubit gates
            elif matrix_type == MatrixType.diagonal:
                value = gate.matrix[_0_1, _0_1]
                Multiply(value) | splited_cgate
            # swap 1-qubit gates [x]
            elif matrix_type == MatrixType.swap:
                DataSwitch(destination, DataSwitchType.all) | splited_cgate
            # reverse 1-qubit gates
            elif matrix_type == MatrixType.reverse:
                value = gate.matrix[int(not _0_1), _0_1]
                Multiply(value) | splited_cgate
                DataSwitch(destination, DataSwitchType.all) | splited_cgate
            # Controlled 1-qubit gates [Z, U1, T, T_dagger, S, S_dagger]
            elif matrix_type == MatrixType.control:
                if _0_1:
                    Multiply(gate.matrix[1, 1]) | splited_cgate
            else:
                SpecialGate(gate.type, outer_qubits_indexes[1], split_idx) | splited_cgate

            dev_mapping[index] = splited_cgate

        return DeviceTrigger(dev_mapping)

    def _double_qubits_gate_transpile(
        self,
        gate: BasicGate,
        inner_args: list,
        outer_args: list
    ):
        """ Deal with double-qubits gate which in the outer qubit.

        Args:
            gate (BasicGate): The quantum gate
            inner_args (list): The gate's qubit indexes which in the inner qubits
            outer_args (list): The gate's qubit indexes which in the outer qubits

        Returns:
            _type_: _description_
        """
        t_qubits = self._qubits - self._split_qb_num
        gate_args, matrix_type = gate.cargs + gate.targs, gate.matrix_type
        double_exceed = len(outer_args) == 2
        if len(outer_args) == 2:
            outer_qubit_idx = [1 << idx for idx, _ in outer_args]
        else:
            intersect_qubit_idx = gate_args.index(outer_args[0][1])
            outer_qubit_idx = 1 << outer_args[0][0]
            inner_args = inner_args[0]

        dev_mapping = {}
        for index in range(self._ndev):
            splited_cgate = CompositeGate()
            # 2-qubits diagonal gate [1, 1, a, b]
            if matrix_type == MatrixType.diagonal:
                if double_exceed:   # both args in splited
                    if index & outer_qubit_idx[0]:
                        value = gate.matrix[3, 3] if index & outer_qubit_idx[1] else gate.matrix[2, 2]
                        Multiply(value) | splited_cgate
                elif intersect_qubit_idx == 0:  # carg in splited
                    if index & outer_qubit_idx:
                        Unitary(gate.target_matrix, MatrixType.diagonal) | splited_cgate(inner_args)
                else:
                    value = gate.matrix[3, 3] if index & outer_qubit_idx else gate.matrix[2, 2]
                    control_matrix = np.array([[1, 0], [0, value]], dtype=np.complex128)
                    Unitary(control_matrix, MatrixType.control) | splited_cgate(inner_args)
            # controlled 2-qubits gate [1, 1, 1, a]
            elif matrix_type == MatrixType.control:
                value = gate.matrix[3, 3]
                control_matrix = np.array([[1, 0], [0, value]], dtype=np.complex128)
                if double_exceed:
                    if (index & outer_qubit_idx[0]) and (index & outer_qubit_idx[1]):
                        Multiply(value) | splited_cgate
                else:
                    if index & outer_qubit_idx:
                        Unitary(control_matrix, MatrixType.control) | splited_cgate(inner_args)
            # diagonal 2-qubits gate [a, b, c, d]
            elif matrix_type == MatrixType.diag_diag:
                if double_exceed:
                    control_index = 2 if index & outer_qubit_idx[0] else 0
                    if index & outer_qubit_idx[1]:
                        control_index += 1

                    value = gate.matrix[control_index, control_index]
                    Multiply(value) | splited_cgate
                elif intersect_qubit_idx == 0:
                    matrix = gate.matrix[2:, 2:] if index & outer_qubit_idx else gate.matrix[:2, :2]
                    Unitary(matrix, MatrixType.diagonal) | splited_cgate(inner_args)
                else:
                    _1 = 1 if index & outer_qubit_idx else 0
                    matrix = np.identity(2, dtype=np.complex128)
                    matrix[[0, 1], [0, 1]] = gate.matrix[[_1, _1 + 2], [_1, _1 + 2]]
                    Unitary(matrix) | splited_cgate(inner_args)
            # control-reverse 2-qubits gate [1, 1, b, a]
            elif matrix_type == MatrixType.reverse:
                if double_exceed:
                    if index & outer_qubit_idx[0]:
                        value = gate.matrix[2, 3] if index & outer_qubit_idx[1] else gate.matrix[3, 2]
                        Multiply(value) | splited_cgate
                        DataSwitch(index ^ outer_qubit_idx[1], DataSwitchType.all) | splited_cgate
                elif intersect_qubit_idx == 0:
                    if index & outer_qubit_idx:
                        matrix = gate.matrix[2:, 2:]
                        Unitary(matrix, MatrixType.reverse) | splited_cgate(inner_args)
                else:
                    matrix = np.identity(2, dtype=np.complex128)
                    matrix[1, 1] = gate.matrix[2, 3] if index & outer_qubit_idx else gate.matrix[3, 2]
                    Unitary(matrix) | splited_cgate(inner_args)
                    DataSwitch(
                        index ^ outer_qubit_idx,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: 1}
                    ) | splited_cgate
            # control-normal 2-qubits gate [CH, CU3]
            elif matrix_type == MatrixType.normal:
                if double_exceed:
                    if index & outer_qubit_idx[0]:
                        DataSwitch(index ^ outer_qubit_idx[1], DataSwitchType.half) | splited_cgate
                        Unitary(gate.matrix[2:, 2:]) | splited_cgate(0)
                        DataSwitch(index ^ outer_qubit_idx[1], DataSwitchType.half) | splited_cgate
                elif intersect_qubit_idx == 0:
                    if index & outer_qubit_idx:
                        Unitary(gate.matrix[2:, 2:]) | splited_cgate(inner_args)
                else:
                    dest = index ^ outer_qubit_idx
                    DataSwitch(
                        dest,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: int(index < dest)}
                    ) | splited_cgate
                    if index & outer_qubit_idx:
                        Unitary(gate.matrix[2:, 2:]) | splited_cgate(inner_args)
                    DataSwitch(
                        dest,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: int(index < dest)}
                    ) | splited_cgate
            # Swap gate [swap]
            elif matrix_type == MatrixType.swap:
                if double_exceed:
                    control_index = 2 if index & outer_qubit_idx[0] else 0
                    if index & outer_qubit_idx[1]:
                        control_index += 1

                    if control_index in [1, 2]:
                        destination = index ^ (sum(outer_qubit_idx))
                        DataSwitch(destination, DataSwitchType.all)
                else:
                    dest = index ^ outer_qubit_idx
                    DataSwitch(
                        dest,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: int(index < dest)}
                    ) | splited_cgate
            # Diagonal * Matrix [Fsim]
            elif matrix_type == MatrixType.ctrl_normal:
                if double_exceed:
                    if index & outer_qubit_idx[0] and index & outer_qubit_idx[1]:
                        Multiply(gate.matrix[-1, -1]) | splited_cgate
                    elif index & outer_qubit_idx[0] or index & outer_qubit_idx[1]:
                        destination = index ^ (sum(outer_qubit_idx))
                        DataSwitch(destination, DataSwitchType.half) | splited_cgate
                        Unitary(gate.matrix[2:, 2:]) | splited_cgate(0)
                        DataSwitch(destination, DataSwitchType.half) | splited_cgate
                else:
                    dest = index ^ outer_qubit_idx
                    DataSwitch(
                        dest,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: 0}
                    ) | splited_cgate
                    if index & outer_qubit_idx:
                        matrix = np.identity(2, dtype=np.complex128)
                        matrix[1, 1] = gate.matrix[3, 3]
                        Unitary(matrix, MatrixType.control) | splited_cgate(inner_args)
                    else:
                        Unitary(gate.matrix[1:3, 1:3]) | splited_cgate(inner_args)
                    DataSwitch(
                        dest,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: 0}
                    ) | splited_cgate
            # Matrix * Matrix [Rxx, Ryy]
            elif matrix_type == MatrixType.normal_normal:
                outside_matrix = gate.matrix[[0, 0, 3, 3], [0, 3, 0, 3]].reshape(2, 2)
                inside_matrix = gate.matrix[1:3, 1:3]
                if double_exceed:
                    destination = index ^ (sum(outer_qubit_idx))
                    DataSwitch(destination, DataSwitchType.half) | splited_cgate
                    matrix = outside_matrix if abs(index - destination) < sum(outer_qubit_idx) else inside_matrix
                    Unitary(matrix) | splited_cgate(0)
                    DataSwitch(destination, DataSwitchType.half) | splited_cgate
                else:
                    dest = index ^ outer_qubit_idx
                    DataSwitch(
                        dest,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: 1}
                    ) | splited_cgate

                    matrix = inside_matrix if index & outer_qubit_idx else outside_matrix
                    Unitary(matrix) | splited_cgate(inner_args)
                    DataSwitch(
                        dest,
                        DataSwitchType.ctarg,
                        {t_qubits - 1 - inner_args: 1}
                    ) | splited_cgate

            dev_mapping[index] = splited_cgate

        return DeviceTrigger(dev_mapping)

    def _gate_switch_cost_generator(self, gates_by_depth: list) -> np.ndarray:
        """ Depending on the gates' type, calculated the switched cost for each gates.

        Args:
            gates_by_depth (list): The list of gates at same layers in circuit.

        Returns:
            ndarray: The switch cost table contains the switched cost of each gates, shape is [# of qubits, # of layers]
        """
        data_switch_cost_matrix = np.zeros((self._qubits, self._depth), dtype=np.int32)
        for idx, depth_gates in enumerate(gates_by_depth):
            for gate in depth_gates:
                gate_matrix_type = gate.matrix_type
                gate_args = gate.cargs + gate.targs
                if gate_matrix_type == MatrixType.normal:
                    if gate.controls == 0:
                        data_switch_cost_matrix[gate_args, idx] += HALF_PENTY
                    else:
                        data_switch_cost_matrix[gate.targs, idx] += CARG_PENTY
                elif gate_matrix_type == MatrixType.reverse:
                    if gate.controls == 1:
                        data_switch_cost_matrix[gate.targs, idx] += CARG_PENTY
                elif gate_matrix_type == MatrixType.swap:
                    if gate.targets > 1:
                        data_switch_cost_matrix[gate_args[-1], idx] += CARG_PENTY
                elif gate_matrix_type in [MatrixType.ctrl_normal, MatrixType.normal_normal]:
                    data_switch_cost_matrix[gate_args, idx] += CARG_PENTY
                elif gate.type in [GateType.measure, GateType.reset]:
                    data_switch_cost_matrix[gate_args, idx] += PROB_ADD

        return data_switch_cost_matrix

    def _split_qubits(self, cost_matrix: np.ndarray):
        """ Depending on the switch cost table, selected the best way to minimize the cost of
        switch data in the distributed environment.

        Args:
            cost_matrix (np.ndarray): The switch cost table contains the switched cost of each gates,
                shape is [# of qubits, # of layers]
        """
        # divided the cost matrix by longest period with switched cost less than cost_threshold
        qubit_num, gate_depth = cost_matrix.shape
        period_interval = []                # record cost matrix's period interval
        sum_cost_by_longest_period = []     # record the sum of period interval of cost matrix
        period_idx = 0
        while period_idx < gate_depth - 1:
            longest_period_idx = 0
            for qubit in range(qubit_num):
                switch_cost = 0
                for idx in range(period_idx, gate_depth, 1):
                    switch_cost += cost_matrix[qubit, idx]
                    if switch_cost > self._cost_threshold:
                        break

                if idx > longest_period_idx:
                    longest_period_idx = idx

            switch_cost = np.sum(cost_matrix[:, period_idx:longest_period_idx], axis=1)
            sum_cost_by_longest_period.append(switch_cost)

            period_idx = longest_period_idx
            period_interval.append(period_idx)

        if gate_depth == 1:
            period_interval.append(gate_depth)
            sum_cost_by_longest_period.append(np.sum(cost_matrix, axis=1))
        else:
            period_interval[-1] += 1

        # find the smallest cost through the sum_cost_by_longest_period
        sum_cost_by_longest_period = np.array(sum_cost_by_longest_period, dtype=np.int32)
        period_qubit_selection = [int(np.argmin(sum_cost_by_longest_period[0]))]
        for sum_cost in sum_cost_by_longest_period[1:]:
            current_minarg = int(np.argmin(sum_cost))
            if current_minarg != period_qubit_selection[-1]:      # required switch split qubits
                # deal with add switch cost threshold
                if sum_cost[period_qubit_selection[-1]] < sum_cost[current_minarg] + self._cost_threshold:
                    current_minarg = period_qubit_selection[-1]
                # deal with switch back cost
                elif len(period_qubit_selection) >= 2 and period_qubit_selection[-2] == current_minarg:
                    start_idx = len(period_qubit_selection) - 2
                    previous_cost = np.sum(sum_cost_by_longest_period[start_idx:start_idx + 2, current_minarg])
                    current_cost = np.sum(
                        sum_cost_by_longest_period[start_idx:start_idx + 2, period_qubit_selection[-2:]]
                    ) + 2 * self._cost_threshold
                    if previous_cost <= current_cost:
                        period_qubit_selection[-1] = current_minarg

            period_qubit_selection.append(current_minarg)

        return period_interval, period_qubit_selection

    def run(self, circuit: Circuit):
        """ Depending on the device number in Transpile, generated the distributed circuit which
        have the necessary DataSwitch and DeviceTrigger operator.

        Args:
            circuit (Circuit): The quantum circuit

        Returns:
            Circuit, List[int]: The distributed circuit and split qubit indexes
        """
        # step 1: run GateDecomposition to avoid Unitary with large qubits and other special gates
        circuit.gate_decomposition()
        self._qubits = circuit.width()
        self._depth = circuit.depth()

        # step 2: order the circuit's gates by depth
        depth_gate = circuit.get_gates_order_by_depth()

        # step 3: decided split qubits by depth gates
        cost_matrix = self._gate_switch_cost_generator(depth_gate)
        split_qubit_intervals, selected_qubit_per_intervals = self._split_qubits(cost_matrix)

        # step 4: transpile circuit by split qubits [add op.data_swap, change gate]
        transpiled_circuits = self._transpile(depth_gate, selected_qubit_per_intervals, split_qubit_intervals)

        return transpiled_circuits, selected_qubit_per_intervals[-1]



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/12/27 5:03 下午
# @Author  : Han Yu
# @File    : __init__.py

from .layout import Layout, LayoutEdge
from .supremacy_layout import SupremacyLayout



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/3/15 1:38
# @Author  : Han Yu
# @File    : _basicInterface.py

from QuICT.core import Circuit


class BasicInterface(object):
    """ the basic interface of general processor

    these interface is devoted to make out circuit more general,
    for example, mutual conversion with qiskit code or OpenQASM 2.0 code.

    the function "load_circuit" and "load_file" should be overloaded

    Attributes:
        circuit(Circuit): the circuit in our framework
        text_content: another form content

    """

    @property
    def circuit(self):
        return self.__circuit

    @circuit.setter
    def circuit(self, circuit):
        self.__circuit = circuit

    @property
    def text_content(self):
        return self.__text_content

    @text_content.setter
    def text_content(self, text_content):
        self.__text_content = text_content

    def __init__(self):
        self.circuit = None
        self.text_content = None

    @staticmethod
    def load_circuit(circuit: Circuit):
        """ load the circuit from our framework

        Args:
            circuit(Circuit): the circuit to be loaded

        """
        instance = BasicInterface()
        instance.circuit = circuit
        return instance

    @staticmethod
    def load_file(filename: str):
        """ load the content from our file

        Args:
            filename(str): filename

        """
        instance = BasicInterface()
        with open(filename) as file:
            instance.text_content = file.read()
        return instance
